package main

import (
	"fmt"
	"os"
	"path"
	"path/filepath"

	"github.com/jessevdk/go-flags"
)

type commandLineOptions struct {
	Config       string `short:"c" long:"config" description:"file with configuration values"`
	Path         string `short:"p" long:"path" description:"define the path where the specifications are located" default:"."`
	DateFormat   string `long:"date-format" description:"date format" default:"2006-01-02"`
	Requirements struct {
		Print struct {
			Columns []string `short:"c" long:"column" description:"defines the columns used"`
		} `command:"print"`
		Add struct {
			ID string `long:"id" required:"true" description:"new requirement id"`
		} `command:"add"`
		Report struct {
			GroupedBy string `long:"grouped-by" required:"false" description:"grouped by"`
			Gitlab    bool   `long:"gitlab" description:"get data from gitlab"`
		} `command:"report"`
	} `command:"requirements"`
	Tasks struct {
		Print struct {
		} `command:"print"`
		Gitlab struct {
			ProjectID string `long:"project-id" required:"false" description:"gitlab project id"`
			Token     string `long:"token" required:"false" description:"gitlab token"`
		} `command:"gitlab"`
	} `command:"tasks"`
	Issues struct {
		Print struct {
		} `command:"print"`
		Gitlab struct {
			ProjectID string `long:"project-id" required:"false" description:"gitlab project id"`
			Token     string `long:"token" required:"false" description:"gitlab token"`
		} `command:"gitlab"`
	} `command:"issues"`
	Items struct {
		Print struct {
		} `command:"print"`
	} `command:"items"`
	Changelog struct {
		Print struct {
		} `command:"print"`
	} `command:"changelog"`
	Privacy struct {
		Print struct {
		} `command:"print"`
	} `command:"privacy"`
	Overview struct {
		Print struct {
			Columns           []string `short:"c" long:"column" description:"defines the columns used"`
			TemplatePath      string   `short:"t" long:"template" description:"file name of the template of the overview page"`
			Output            string   `short:"o" long:"output" description:"redirect output to a file"`
			Format            string   `short:"f" long:"format" description:"the desired output format" choice:"pdf" choice:"md"`
			LatexTemplatePath string   `short:"l" long:"latex-template-path" description:"latex template"`
		} `command:"print"`
	} `command:"overview"`
	Version struct {
	} `command:"version"`
}

func executeCommand() {

	arguments := new(commandLineOptions)
	p := flags.NewParser(arguments, flags.Default)
	_, err := p.Parse()
	if err != nil {
		os.Exit(-1)
	}

	config := NewConfiguration(arguments.Config)

	if arguments.Path != "" {
		config.Path = arguments.Path
	}

	if arguments.DateFormat != "" {
		config.Locale.DateFormat = arguments.DateFormat
	}

	activeCommand := p.Command.Active

	if activeCommand == nil {
		return
	}

	switch activeCommand.Name {

	case "version":
		fmt.Println("Version " + version + " (" + build + ")")
	case "changelog":
		subcommand := activeCommand.Active
		switch subcommand.Name {
		case "print":
			err := printChangelog(config)
			if err != nil {
				printErrorAndExit(2, err.Error())
			}
		}
	case "privacy":
		subcommand := activeCommand.Active
		switch subcommand.Name {
		case "print":
			err := printPrivacy(config)
			if err != nil {
				printErrorAndExit(2, err.Error())
			}
		}
	case "overview":
		subcommand := activeCommand.Active
		switch subcommand.Name {
		case "print":

			if len(arguments.Overview.Print.Columns) > 0 {
				config.Requirement.Table.Columns = validateColumns(arguments.Overview.Print.Columns)
			}

			if arguments.Overview.Print.TemplatePath != "" {
				config.Overview.Template.Internal.MarkdownContent = readTemplate(arguments.Overview.Print.TemplatePath)
			} else if config.Overview.Template.Markdown != "" {
				config.Overview.Template.Internal.MarkdownContent = readTemplate(config.Overview.Template.Markdown)
			} else {
				config.Overview.Template.Internal.MarkdownContent = defaultOverviewTemplate
			}

			if arguments.Overview.Print.LatexTemplatePath != "" {
				config.Overview.Template.Latex = arguments.Overview.Print.LatexTemplatePath
			}

			err := PrintOverview(config, arguments)
			if err != nil {
				printErrorAndExit(2, err.Error())
			}
		}

	case "requirements":
		subcommand := activeCommand.Active
		switch subcommand.Name {
		case "report":
			err := printRequirementReport(config, arguments)
			if err != nil {
				printErrorAndExit(2, err.Error())
			}
		case "add":
			config.Requirement.Template.Add = defaultNewRequirementTemplate
			err := addRequirement(config, arguments)
			if err != nil {
				printErrorAndExit(2, err.Error())
			}
		case "print":
			if len(arguments.Requirements.Print.Columns) > 0 {
				config.Requirement.Table.Columns = validateColumns(arguments.Requirements.Print.Columns)
			}
			err := printRequirements(config)
			if err != nil {
				printErrorAndExit(2, err.Error())
			}
		}

	case "tasks":
		subcommand := activeCommand.Active
		switch subcommand.Name {
		case "print":
			err := printTaskTable(config)
			if err != nil {
				printErrorAndExit(2, err.Error())
			}
		}

	case "issues":
		subcommand := activeCommand.Active
		switch subcommand.Name {
		case "print":
			err := printIssueTable(config)
			if err != nil {
				printErrorAndExit(2, err.Error())
			}
		case "gitlab":
			syncTasksWithGitlab(config)
		}

	case "items":
		subcommand := activeCommand.Active
		switch subcommand.Name {
		case "print":
			err := printItemTable(config)
			if err != nil {
				printErrorAndExit(2, err.Error())
			}
		}

	}
}

func readTemplate(argPath string) string {

	current, err := os.Getwd()
	if err != nil {
		printErrorAndExit(2, "Unknown Error", err.Error())
	}

	p := argPath

	if !filepath.IsAbs(p) {
		p = path.Clean(current + "/" + p)
	}

	template, err := os.ReadFile(p)
	if err != nil {
		printErrorAndExit(exitCodeCatchAll,
			"the specified template", argPath)
	}

	return convertTemplateImages(string(template), path.Dir(p))
}
