package main

import (
	"net/url"
	"path"

	"github.com/go-git/go-git/v5"
	"github.com/xanzy/go-gitlab"
)

type gitlabContext struct {
	rootPath string
	repos    *git.Repository
	apiUrl   url.URL

	group string
	name  string

	token  string
	client *gitlab.Client

	project *gitlab.Project
}

func getGitlabContext(config *Configuration) *gitlabContext {

	context := &gitlabContext{}
	workingPath, err := getGitDirectory(config)
	if err != nil {
		printErrorAndExit(2, "Failed to get git directory", err.Error())
	}

	context.rootPath = workingPath

	repos, err := git.PlainOpen(workingPath)
	if err != nil {
		printErrorAndExit(2, "Failed to open repository", err.Error())
	}

	context.repos = repos
	if err != nil {
		printErrorAndExit(2, "Failed to get remotes", err.Error())
	}

	remotes, err := repos.Remotes()
	if err != nil {
		printErrorAndExit(2, "Failed to get remotes", err.Error())
	}

	remote := remotes[0]
	if remote == nil {
		printErrorAndExit(2, "Failed to get remote")
	}

	urls := remote.Config().URLs
	if len(urls) == 0 {
		printErrorAndExit(2, "Failed to get remote url")
	}

	u, err := url.Parse(urls[0])
	if err != nil {
		printErrorAndExit(2, "Failed to parse remote url", err.Error())
	}

	context.apiUrl = *u
	context.apiUrl.Path = "" //path.Join("api", "v4")

	p := u.Path
	if p == "" {
		printErrorAndExit(2, "Failed to get remote path")
	}

	s := p[1 : len(p)-4]
	context.name = path.Base(s)
	context.group = path.Dir(s)

	options := gitlab.WithBaseURL(context.apiUrl.String())

	context.token = config.GitLab.Token
	c, err := gitlab.NewClient(config.GitLab.Token, options)
	if err != nil {
		printErrorAndExit(2, "Failed to create client", err.Error())
	}

	context.client = c

	return context
}

func searchProject(context *gitlabContext) {
	//
	options := gitlab.ListProjectsOptions{
		Search:           gitlab.String(context.group + "/" + context.name),
		SearchNamespaces: gitlab.Bool(true),
	}

	projects, response, err := context.client.Projects.ListProjects(&options)

	if response.StatusCode == 404 {
		printErrorAndExit(2, "Project not found %s", err.Error())
	}

	if err != nil {
		printErrorAndExit(2, "Failed to get projects %s", err.Error())
	}

	if response.StatusCode != 200 {
		printErrorAndExit(2, "Failed to get projects %s", err.Error())
	}

	if len(projects) == 0 {
		printErrorAndExit(2, "Project not found")
	}

	context.project = projects[0]

}

func enrichIssuesWithGitlab(pageMap *map[string]requirement, config *Configuration) error {

	context := getGitlabContext(config)
	searchProject(context)

	for _, requirement := range *pageMap {
		for k, issue := range requirement.Issues {

			if issue.GitlabID == 0 {
				continue
			}

			i, err := loadIssuesFromGitlab(context, issue.GitlabID)
			if err != nil {
				return err
			}

			// requirement is a copy of the original requirement
			// this works because issue is a pointer
			requirement.Issues[k].Gitlab = i
		}
	}

	return nil
}

//
//func getProject(config *Configuration) *gitlab.Project {
//
//	client := getGitlabClient(config)
//
//	pattern := "test"
//
//	options := gitlab.ListProjectsOptions{
//		Search: &pattern,
//	}
//
//	projects, _, err := client.Projects.ListProjects(&options)
//	if err != nil {
//		printErrorAndExit(2, "Failed to get projects", err.Error())
//	}
//
//	//project, _, err := client.Projects.GetProject(config.GitlabProjectID, nil)
//	//if err != nil {
//	//	printErrorAndExit(2, "Failed to get project", err.Error())
//	//}
//
//	return projects[0]
//}
