package main

import (
	"bytes"
	"fmt"
	"math"
	"os"
	"path"
	"reflect"
	"strconv"
	"strings"
	"time"

	"github.com/olekukonko/tablewriter"
)

type requirement struct {
	ToDos    []task
	Absolute string
	File     string
	Items    []Item    `yaml:"Items"`
	Privacy  []Privacy `yaml:"Privacy"`
	Issues   []Issue   `yaml:"Issues"`

	ID         string        `yaml:"ID"`
	References []string      `yaml:"References"`
	Type       string        `yaml:"Type"` // Display, Functional, Performance, Printing, Report, Testing and Validate
	Title      string        `yaml:"Title"`
	Alias      []string      `yaml:"Alias"`
	Keywords   []string      `yaml:"Keywords"`
	Author     []string      `yaml:"Author"`
	Status     string        `yaml:"Status"`
	Complexity string        `yaml:"Complexity"`
	Difficulty string        `yaml:"Difficulty"` // Low, Medium, High
	Priority   string        `yaml:"Priority"`   // Low, Medium, High
	Version    string        `yaml:"Version"`
	Milestone  string        `yaml:"Milestone"`
	Estimation int           `yaml:"Estimation"` // 0, 1, 2, 3, 5, 8, 13, 20, 40, 100
	TimeSpent  time.Duration `yaml:"Time Spent"` // 3h2m20s
	Source     string        `yaml:"Source"`
	Created    time.Time     `yaml:"Created"`
	LastUpdate time.Time     `yaml:"Last Update"`
}

func addRequirement(config *Configuration, arguments *commandLineOptions) error {

	if arguments.Requirements.Add.ID == "" {
		printErrorAndExit(2, "the id must not be empty")
	}

	p := path.Join(arguments.Path, arguments.Requirements.Add.ID+".md")
	if fileExists(p) {
		printErrorAndExit(2, "the request with id already exists", arguments.Requirements.Add.ID, p)
	}

	t := config.Requirement.Template.Add
	t = strings.Replace(t, "%%ID%%", arguments.Requirements.Add.ID, -1)
	t = strings.Replace(t, "%%CREATED%%", time.Now().Format("2006-01-02"), -1)

	d1 := []byte(t)
	err := os.WriteFile(p, d1, 0644)
	if err != nil {
		return err
	}

	return nil
}

func buildRequirements(config *Configuration, pageMap map[string]requirement) (error, string, bool) {

	hasRequirements := false

	buf := new(bytes.Buffer)
	table := tablewriter.NewWriter(buf)

	translateMetaColumns := translateHeaders(config.Requirement.Table.Columns)

	table.SetHeader(translateMetaColumns)
	table.SetBorders(tablewriter.Border{Left: true, Top: false, Right: true, Bottom: false})
	table.SetCenterSeparator("|")

	var tableData [][]string

	for _, pd := range pageMap {

		r := reflect.ValueOf(pd)

		var row []string
		found := false
		for _, name := range config.Requirement.Table.Columns {

			field := reflect.Indirect(r).FieldByName(name)
			if field == (reflect.Value{}) {
				row = append(row, printer.Sprintf("no value"))
				continue
			}

			t := field.Type()

			switch t.Name() {
			case "string":
				found = true
				row = append(row, field.String())
			case "Duration":
				found = true
				t := field.Interface().(time.Duration)
				row = append(row, t.String())
			case "Time":
				found = true
				t := field.Interface().(time.Time)
				if t.IsZero() {
					row = append(row, "—")
				} else {
					row = append(row, field.Interface().(time.Time).Format(config.Locale.DateFormat))
				}
			case "int":
				found = true
				row = append(row, strconv.FormatInt(field.Int(), 10))
			case "float64":
				found = true
				row = append(row, strconv.FormatFloat(field.Float(), 'f', 10, 2))
			case "nil":
				row = append(row, field.String())
			case "bool":
				found = true
				expr := field.Bool()
				if expr {
					row = append(row, "true")
				} else {
					row = append(row, "false")
				}

			default:
				row = append(row, printer.Sprintf("unsuported type", t.Name()))
			}

		}

		if found {
			tableData = append(tableData, row)
			hasRequirements = true
		}

	}

	table.AppendBulk(tableData) // Add Bulk Data
	table.SetHeaderAlignment(tablewriter.ALIGN_LEFT)
	table.Render()

	return nil, buf.String(), hasRequirements

}

func printRequirements(config *Configuration) error {
	err, pageData := collectStructureFromFiles(config.Path)
	if err != nil {
		return err
	}

	err, table, _ := buildRequirements(config, pageData)
	fmt.Println(table)
	return err
}

func integrateGitlabTimeSpent(pageMap *map[string]requirement) {

	for k, requirement := range *pageMap {

		for _, i := range requirement.Issues {
			if i.Gitlab == nil {
				continue
			}

			//d := time.Duration(i.Gitlab.TimeStats.TotalTimeSpent) * time.Second
			//fmt.Println(fn, d)
			//ref[fn].TimeSpent += d
			requirement.TimeSpent = 99999999
			requirement.Status = "done"

			//			fmt.Println(i.Gitlab.TimeStats.TotalTimeSpent)
		}

		pageMap[k] = requirement

		//if &pd.Issues == nil || &pd.Issues. == nil {
		//	continue
		//}
		//
		//for _, issue := range pd.Issues.Gitlab {
		//	if issue == nil {
		//		continue
		//	}
		//	pd.TimeSpent += *issue.TimeSpent
		//}

	}
	fmt.Println("integrateGitlabTimeSpent")

}

func buildTimeReport(config *Configuration, arguments *commandLineOptions, pageMap *map[string]requirement) (error, string, string, bool) {

	if arguments.Requirements.Report.Gitlab {
		err := enrichIssuesWithGitlab(pageMap, config)
		if err != nil {
			return err, "", "", false
		}

		integrateGitlabTimeSpent(pageMap)

	}

	groupedByKey := arguments.Requirements.Report.GroupedBy
	if groupedByKey == "" {
		groupedByKey = "ID"
	}

	hasReport := false

	found := false
	for _, p := range *pageMap {

		fields := reflect.VisibleFields(reflect.TypeOf(p))
		for _, field := range fields {
			if field.Name == groupedByKey {
				found = true
				break
			}
		}
		if found {
			break
		}
	}

	if !found {
		printErrorAndExit(2, "the grouping key was not found", groupedByKey)
	}

	t1 := buildTable1(groupedByKey, pageMap)

	t2 := buildTable2(pageMap)

	return nil, t1, t2, hasReport
}

func buildTable1(groupedByKey string, pageMap *map[string]requirement) string {

	estimated := map[string]float64{}
	timeSpent := map[string]float64{}

	for _, d := range *pageMap {

		r := reflect.ValueOf(d)
		f := r.FieldByName(groupedByKey)
		k := ""
		switch f.Interface().(type) {
		case bool:
			k = strconv.FormatBool(r.FieldByName(groupedByKey).Bool())
		case int:
			k = strconv.FormatInt(r.FieldByName(groupedByKey).Int(), 10)
		case int64:
			k = strconv.FormatInt(r.FieldByName(groupedByKey).Int(), 10)
		case float64:
			k = strconv.FormatFloat(r.FieldByName(groupedByKey).Float(), 'f', 10, 2)
		case string:
			k = r.FieldByName(groupedByKey).String()
		default:
			k = r.FieldByName(groupedByKey).String()
		}

		s := float64(d.TimeSpent.Milliseconds()) / 1000

		estimated[k] += float64(d.Estimation)
		timeSpent[k] += s

	}

	buf := new(bytes.Buffer)
	table := tablewriter.NewWriter(buf)

	columns := []string{"Key", "Estimated", "Time Spent", "Ratio"}

	table.SetHeader(translateHeaders(columns))
	table.SetBorders(tablewriter.Border{Left: true, Top: false, Right: true, Bottom: false})
	table.SetCenterSeparator("|")

	var tableData [][]string

	for k := range estimated {
		var row []string
		row = append(row, k)
		row = append(row, strconv.FormatInt(int64(estimated[k]), 10))
		// nanosecond
		row = append(row, time.Duration(timeSpent[k]*1000000000).String())

		var ratio float64
		ratio = 0
		if timeSpent[k] > 0 {

			e := float64(estimated[k]) * 3600
			s := timeSpent[k]
			ratio = s / e * 100

		}

		row = append(row, fmt.Sprintf("%.2f", ratio))
		tableData = append(tableData, row)
	}

	table.AppendBulk(tableData) // Add Bulk Data
	table.SetHeaderAlignment(tablewriter.ALIGN_LEFT)
	table.Render()

	return buf.String()
}

type timeStruct struct {
	sum         float64
	hour        map[int]int // [time]count
	average     float64
	sqrtAverage float64
}

type mathData struct {
	count    float64
	estimate struct {
	}
	time timeStruct
}

func buildTable2(pageMap *map[string]requirement) string {

	est := map[int]mathData{}

	// inti struct
	for _, d := range *pageMap {
		k := d.Estimation

		m := mathData{}
		m.count = 0
		m.time = timeStruct{
			0,
			make(map[int]int),
			0,
			0,
		}

		est[k] = m
	}

	// calc sum
	for _, d := range *pageMap {
		k := d.Estimation
		s := est[k]

		s.count++
		sec := float64(d.TimeSpent.Milliseconds()) / 1000
		s.time.sum += sec

		h := int(math.Round(sec / 3600))
		mh := s.time.hour
		mh[h]++

		est[k] = s
	}

	for k, s := range est {
		//
		a := 0.0
		c := 0.0

		for h, m := range s.time.hour {
			a += float64(h * m)
			//			g += math.Pow(a, 2)
			c += float64(m)
		}

		x := est[k]
		x.time.average = a / c

		g := 0.0
		for q, m := range s.time.hour {
			z := float64(q) - x.time.average
			g += math.Pow(z, 2) * float64(m)
		}

		x.time.sqrtAverage = g / c
		est[k] = x

	}

	buf := new(bytes.Buffer)
	table := tablewriter.NewWriter(buf)

	columns := []string{"Estimated", "Time Spent", "Average", "Variance", "Standard Deviation"}

	table.SetHeader(translateHeaders(columns))
	table.SetBorders(tablewriter.Border{Left: true, Top: false, Right: true, Bottom: false})
	table.SetCenterSeparator("|")

	var tableData [][]string

	for e, c := range est {

		row := []string{}
		row = append(row, fmt.Sprintf("%d", e))
		row = append(row, time.Duration(c.time.sum*float64(time.Second)).String())
		row = append(row, fmt.Sprintf("%.2f", c.time.average))
		row = append(row, fmt.Sprintf("%.2f", c.time.sqrtAverage))
		row = append(row, fmt.Sprintf("%.2f", math.Sqrt(c.time.sqrtAverage)))

		tableData = append(tableData, row)

	}

	table.AppendBulk(tableData) // Add Bulk Data
	table.SetHeaderAlignment(tablewriter.ALIGN_LEFT)
	table.Render()

	return buf.String()
}

func printRequirementReport(config *Configuration, arguments *commandLineOptions) error {

	err, pageData := collectStructureFromFiles(config.Path)
	if err != nil {
		return err
	}

	err, table1, table2, _ := buildTimeReport(config, arguments, &pageData)
	if err != nil {
		return err
	}

	report := "## " + printer.Sprintf("Auflistung") + "\n\n"
	report += table1
	report += "\n"
	report += "## " + printer.Sprintf("Statistics") + "\n\n"
	report += table2

	fmt.Println(report)

	return nil
}
